"""
Asset Management System
(C) Copyright 2004 Hamster Republic Productions
"""

import os.path
import sys
import pickle
import types
import weakref
import copy
from motherhamster import enums
from motherhamster import fooddish
from motherhamster.mathobj import vec2d

########################################################################
class AssetBase(object):
    """a base class for assets - all assets should inherit from it"""
    def __init__(self):
        #CHECK: should I make asset_metadata a property descriptor with a get that returns an object we can use . access on?
        self.asset_metadata = {}
    
    def __getinitargs__(self):
        return ()
    
    def __reduce__(self):
        dict = {}
        for class_type in type(self).mro():
            for key, value in class_type.__dict__.items():
                if issubclass(type(value), AssetProperty):
                    if not value.asset_has_default(self, key):
                        dict[key] = self.__dict__[key]
        return (self.__class__, (), dict)

    def asset_get_properties(self):
        """Returns a dictionary of the attributes which are asset properties"""
        properties = {}
        for class_type in type(self).mro():
            for key, value in class_type.__dict__.iteritems():
                if issubclass(type(value), AssetProperty):
                    properties[key] = value
        return properties
        
    def asset_is_writable(self):
        """Returns whether you will be able to write the asset to disk"""
        try:
            path = self.asset_metadata['path']
        except:
            return False
        return GetFoodDish(path.root).is_writable(path)

    def asset_set_path(self, path, pouch):
        if not hasattr(self, "asset_metadata"):
            self.asset_metadata = {}
        self.asset_metadata['path'] = path
        self.asset_metadata['pouch'] = weakref.proxy(pouch)
        for key in self.asset_get_properties():
           if self.__dict__.has_key(key) and issubclass(type(getattr(self.__class__, key)), AssetObject):
               getattr(self, key).asset_set_path(path, pouch)
        
    def asset_save(self):
        """Asks the cheekpouch for this asset to save it, returns False if we are not writable"""
        path = self.asset_metadata['path']
        pouch = self.asset_metadata['pouch']
        assert(self.asset_is_writable())
        pouch._save(path, self)
        return True
    
    def asset_editor_image(self):
        """Returns an image representation of the asset as a pygame surface"""
        return None
    
    def asset_monitor_changes(self, object_to_call):
        """Registers an object to be called when the asset is modified"""
        if (not self.asset_metadata.has_key('monitor_list')):
            self.asset_metadata['monitor_list'] = []
        monitor_list = self.asset_metadata['monitor_list']
        monitor_list.append(weakref.ref(object_to_call, self.asset_lose_monitor))
    
    def asset_lose_monitor(self, object_to_call):
        """removes an object from the registry"""
        monitor_list = self.asset_metadata['monitor_list']
        monitor_list.remove(object_to_call)
    
    def asset_report_property_change(self, key, value):
        """call to tell the asset you changed it, if you want monitors to know about it"""
        if (self.asset_metadata.has_key('monitor_list')):
            monitor_list = self.asset_metadata['monitor_list']
            for object_ref in monitor_list:
                object_to_call = object_ref()
                if object_to_call != None:
                    object_to_call.asset_changed(self, key, value)
                    
    def asset_set_property(self, key, value):
        """A function the editor calls to change a property
           the editor should never edit an asset any other way"""
        property_descriptor = getattr(type(self), key)
        assert(isinstance(property_descriptor, AssetProperty))
        setattr(self, key, value)
        self.asset_report_property_change(key, value)
                
        
########################################################################
class GenericAssetPouch(object):
    """the base class for asset pouches"""
    def __init__(self, classtype, broken_reference = None):
        assert(issubclass(classtype, AssetBase))
        self.classtype = classtype
        self.broken_reference = broken_reference
        self._loaded_assets = {}
        self._asset_paths = None

    def AvailableAssetIDs(self):
        """returns a list of strings for valid asset names"""
        if (self._asset_paths != None):
            return self._asset_paths
        self._asset_paths = []
        for dish in AvailableFoodDishes():
            self._asset_paths.extend(dish.FindFilesByExtension(self.AssetExtensions()))
        return self._asset_paths
    
    def RefreshAssetPaths(self):
        self._asset_paths = None
        
    def AssetExtensions(self):
        """returns a list of supported extensions, lowercase with dot prefix"""
        raise Exception("You need to implement support extensions")

    def NewAssetExtension(self):
        """gets the extension for new assets (should be lowercase and start with .)
           return None if this pouch does not support new assets"""
        return None
    
    def GetAssetPathForID(self, assetid):
        """find the best matching asset path for a given id"""
        for asset_path in self.AvailableAssetIDs():
            if asset_path == assetid:
                return asset_path
        return None

    def DisgorgeAsset(self, assetid):
        """either retrieves the loaded asset or loads it"""
        if self._loaded_assets.has_key(assetid):
            return self._loaded_assets[assetid]
        path = self.GetAssetPathForID(assetid)
        if (path == None):
            fail_message = "Failed to find asset " + str(assetid) + " of type " + str(self.classtype)
            if (self.broken_reference == None):
                raise Exception(fail_message)
            else:
                if ReportBrokenAsset(assetid, self.classtype):
                    print fail_message
                return self.broken_reference
        obj = self._load(path)
        self.AddAssetWithPath(obj, path)
        return obj

    def AddAssetWithPath(self, asset, asset_path):
        self._loaded_assets[asset_path] = asset
        asset.asset_set_path(asset_path, self)
        
    def UpdateAsset(self, assetid):
        """saves the in-memory copy of an asset to persistent storage"""
        path = self.GetAssetPathForID(assetid)
        obj = self._loaded_assets[assetid]
        self._save(path, obj)
        
    def RevertAsset(self, assetid):
        """sets the in-memory copy of an asset to it's persistent storage copy if loaded
           note that this updates the in-memory copy, as opposed to making a new copy"""
        if self._loaded_assets.has_key(assetid):
            path = self.GetAssetPathForID(assetid)
            new_copy = self._load(path)
            self._loaded_assets[assetid].__dict__.update(new_copy.__dict__)
        
    def ClassType(self):
        """gets the type which defines the interface of loaded assets from this pouch
           loaded objects may not be of this type, in a pythonic fashion"""
        return self.classtype
        
    def CreateNewAsset(self, asset_path):
        """ creates a newasset """
        new_obj = self.classtype()
        self.AddAssetWithPath(new_obj, asset_path)
        new_obj.asset_save()
        return new_obj
                
    def _load(self, asset_path):
        raise Exception("You need to implement asset loading - requested " + assetid)

    def _save(self, asset_path, obj):
        raise Exception("You need to implement asset saving - requested " + assetpath)
        
########################################################################
class SimpleAssetPouch(GenericAssetPouch):
    """an asset pouch designed to be instantiated for objects that are saved and loaded by simple pickle"""
    def __init__(self, classtype, extension, broken_reference = None):
        GenericAssetPouch.__init__(self, classtype, broken_reference)
        extension = extension.lower()
        if (extension[0] != '.'):
            extension = '.' + extension        
        self.extension_list = [extension]

    def AssetExtensions(self):
        return self.extension_list
    
    def NewAssetExtension(self):
        return self.extension_list[0]
    
    def _load(self, asset_path):
        """loads an asset using pickle based on the extension"""
        dish = GetFoodDish(asset_path.root)
        asset_file = dish.open(asset_path.filepath, "rb")
        return pickle.load(asset_file)

    def _save(self, asset_path, obj):
        """saves an asset using pickle"""
        dish = GetFoodDish(asset_path.root)
        asset_file = dish.open(asset_path.filepath, "wb")
        pickle.dump(obj, asset_file)

########################################################################
class AssetProperty(object):
    """the base class for an editable property on an asset"""
    
    def _myname(self, obj, objtype):
        """gets the key of the AssetRef descriptor on the class of the object"""
        if hasattr(self, "_cached_name"):
            return self._cached_name
        class_to_check = objtype
        while (class_to_check != None):
            for key, value in class_to_check.__dict__.iteritems():
                if value == self:
                    self._cached_name = key
                    return key
            class_to_check = super(class_to_check, obj)       
        raise Exception("ERROR: AssetProperty couldn't find its name")
    
    def asset_has_default(self, obj, key):
        return not obj.__dict__.has_key(key)
    
########################################################################
class SimpleProperty(AssetProperty):
    """an assetproperty where you just have a simple value stored"""
    def __init__(self, asset_type, default, description = None):
        AssetProperty.__init__(self)
        self.type = asset_type
        self.default = default
        self.description = description
        assert(isinstance(default, asset_type))
        
    def __get__(self, obj, objtype=None):    #IGNORE:W0613
        if obj is None:
            return self         
        return self.default

    def get_new(self):
        return self.default
    
########################################################################
class PropertyDeepCopyDefault(SimpleProperty):
    """an assetproperty where the default value needs a deep copy"""
    
    def __get__(self, obj, objtype=None):    #IGNORE:W0613
        if obj is None:
            return self
        attribute_name = self._myname(obj, objtype)
        default_copy = self.get_new()
        obj.__dict__[attribute_name] = default_copy
        return default_copy
    
    def asset_has_default(self, obj, key):
        return not obj.__dict__.has_key(key) or (obj.__dict__[key] == self.default)

    def get_new(self):
        return copy.deepcopy(self.default)
            
########################################################################
class AssetFlag(SimpleProperty):
    """the class for an editable flag on an asset"""
    def __init__(self, default = False, description = None):
        SimpleProperty.__init__(self, types.BooleanType, default, description)

########################################################################
class AssetInteger(SimpleProperty):
    """the class for an editable number that can take a range of values"""
    def __init__(self, default = 0, units = "", min = -2**31, max = 2**31, description = None, step = None):
        SimpleProperty.__init__(self, types.IntType, default, description)
        assert(isinstance(units, str))
        self.units = units
        self.min = min
        self.max = max
        if (step == None):
            self.step = int((self.max - self.min)/100)
            if self.step > 2**16 or self.step == 0:
                if abs(default) < 10:
                    self.step = 1
                else:
                    self.step = abs(default)/10
        else:
            self.step = step

########################################################################
class AssetFloat(SimpleProperty):
    """the class for an editable number that can take a range of values"""
    def __init__(self, default = 0.0, units = "", min = -1e300, max = 1e300, description = None, step = None):
        SimpleProperty.__init__(self, types.FloatType, default, description)
        assert(isinstance(units, str))
        self.units = units
        self.min = min
        self.max = max
        if (step == None):
            self.step = (self.max - self.min)/100.0
            if self.step > 1e9:
                if default == 0:
                    self.step = 1.0
                else:
                    self.step = abs(default)/10.0
        else:
            self.step = step
        
########################################################################
class AssetString(SimpleProperty):
    """the class for an editable string on an asset"""
    def __init__(self, default = "", description = None):
        SimpleProperty.__init__(self, types.StringType, default, description)
        
########################################################################
class AssetEnum(SimpleProperty):
    """the class for an editable number that can take a range of values"""
    def __init__(self, enum, default = None, description = None):
        if default == None:
            default = enum[0]
        self.enum = enum
        SimpleProperty.__init__(self, str, default, description)

########################################################################
class AssetVec2d(PropertyDeepCopyDefault):        
    """the class for an editable 2d vector"""
    def __init__(self, default = vec2d(0,0), description = None):
        if not isinstance(default, vec2d):
            default = vec2d(default)
        PropertyDeepCopyDefault.__init__(self, vec2d, default, description)
   
########################################################################
class AssetColor(PropertyDeepCopyDefault):        
    """the class for an editable color"""
    def __init__(self, default = [255, 255, 255, 255], description = None):
        PropertyDeepCopyDefault.__init__(self, types.ListType, default, description)
   
########################################################################
class AssetObject(AssetProperty):
    """a reflected object"""
    def __init__(self, base_class = AssetBase, description = None):
        AssetProperty.__init__(self)
        self.description = description
        self.object_class = base_class

    def __get__(self, obj, objtype=None):    #IGNORE:W0613
        if obj is None:
            return self
        attribute_name = self._myname(obj, objtype)
        new_obj = self.object_class()
        new_obj.asset_metadata = obj.asset_metadata
        obj.__dict__[attribute_name] = new_obj
        return new_obj

    def get_new(self):
        return self.object_class()

########################################################################
class AssetList(AssetProperty):
    """a reflected list of reflected somethings"""
    def __init__(self, element_descriptor, description = None):
        AssetProperty.__init__(self)
        self.descriptor = element_descriptor
        assert(not isinstance(element_descriptor, AssetRef))
        self.description = description

    def __get__(self, obj, objtype=None):    #IGNORE:W0613
        if obj is None:
            return self
        attribute_name = self._myname(obj, objtype)
        new_list = []
        obj.__dict__[attribute_name] = new_list
        return new_list
        
########################################################################
class AssetDictionary(AssetProperty):
    """a reflected dictionary of reflected somethings"""
    def __init__(self, element_descriptor, description = None):
        AssetProperty.__init__(self)
        self.descriptor = element_descriptor
        assert(isinstance(element_descriptor, SimpleProperty))
        self.description = description

    def __get__(self, obj, objtype=None):    #IGNORE:W0613
        if obj is None:
            return self
        attribute_name = self._myname(obj, objtype)
        new_dict = {}
        obj.__dict__[attribute_name] = new_dict
        return new_dict
        
########################################################################
class AssetRef(AssetProperty):
    """the only valid way to reference an asset - put one of these as a 
        descriptor on your class for a reference to an asset, switch assets by
        assigning this to the assetid, access the asset through this descriptor
        exclusively - if you don't, your object will serialize incorrectly
        The descriptor will store the assetid in an attribute on the instance,
        using the key you assign the AssetRef descriptor to on the class"""
    def __init__(self, asset_type, default_assetid = "!", description = None):
        AssetProperty.__init__(self)
        self.type = asset_type
        self.default_assetid = default_assetid
        self.description = description
        
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        # CHECK: should we cache the asset instance? probably...
        binding = self._myname(obj, objtype)
        if (obj.__dict__.has_key(binding)):
            assetid = obj.__dict__[binding]
        else:
            assetid = self.default_assetid
        return Disgorge(self.type, assetid)
    
    def __set__(self, obj, value):
        binding = self._myname(obj, type(obj))
        if isinstance(value, AssetBase):
            raise Exception("ERROR: whoops, we don't support setting to assets yet, use the AssetID instead...")
        else:
            obj.__dict__[binding] = value

    def __delete__(self, obj):
        del obj.__dict__[self._myname(obj, type(obj))]
    
    def get_assetid(self, obj):
        key = self._myname(obj, type(obj))
        if not obj.__dict__.has_key(key):
            return self.default_assetid
        return obj.__dict__[key]
        
        
########################################################################

class AssetTypeError(Exception):
    """Exception for when a classtype that has no Disgorgeer is requested"""

########################################################################

_AssetPouchList = {}
_AssetExtensionList = {}
_AssetsNotFoundList = []

def RegisterPouch(pouch):
    """registers the Disgorgeer for its class type"""
    global _AssetPouchList
    classtype = pouch.ClassType()
    assert not _AssetPouchList.has_key(classtype), "Two asset Disgorgeers are advertising the same classtype, not supported"
    _AssetPouchList[classtype] = pouch

def GetPouch(class_type):
    """gets the pouch for a class type - returns None if none registered"""
    global _AssetPouchList
    if _AssetPouchList.has_key(class_type):
        return _AssetPouchList[class_type]
    else:
        return None
    
def RegisterAssetExtension(class_type, extension, broken_reference = None):
    """registers an asset type that doesn't need any special loader behavior"""
    global _AssetExtensionList
    if (extension[0] == "."):
        extension = extension[1:]
    assert not _AssetExtensionList.has_key(extension), "Multiple asset types " + str(class_type) + " & " + str(_AssetExtensionList[extension]) + " registered with extension " + extension
    _AssetExtensionList[extension] = class_type
    RegisterPouch(SimpleAssetPouch(class_type, extension, broken_reference))
    
def AvailableAssetClassTypes():
    """returns a list of base class types for which assets exist"""
    global _AssetPouchList
    return _AssetPouchList.keys()

def AvailableAssetIDs(class_type, rescan_list = False):
    """returns a list of strings which contain asset names"""
    global _AssetPouchList
    if _AssetPouchList.has_key(class_type):
        if rescan_list:
            _AssetPouchList[class_type].RefreshAssetPaths()
        return _AssetPouchList[class_type].AvailableAssetIDs()
    else:
        return None

def Disgorge(class_type, asset_id, verify_assetid = False):
    """given an asset id and class type, returns the asset in a wrapper"""
    global _AssetPouchList
    if _AssetPouchList.has_key(class_type):
        pouch = _AssetPouchList[class_type]
        if (verify_assetid):
            if (not asset_id in pouch.AvailableAssetIDs()):
                raise Exception("Asked for an invalid asset id: "+asset_id)
        return pouch.DisgorgeAsset(asset_id)
    else:
        raise AssetTypeError(str(class_type) + " is not a registered class type\nAvailableList:" + str(AvailableAssetClassTypes()))

_FoodDishList = {}

def ReportBrokenAsset(asset_id, class_type):
    for entry in _AssetsNotFoundList:
        if entry['asset_id'] == asset_id and entry['class'] == class_type:
            return False
    _AssetsNotFoundList.append({'asset_id':asset_id, 'class':class_type})
    return True
    
def RegisterFoodDish(dish):        
    _FoodDishList[dish.name] = dish

def AvailableFoodDishes():
    return _FoodDishList.values()

def GetFoodDish(name):
    return _FoodDishList[name]

#TEMP: register the local data path - should figure out who should own this...
RegisterFoodDish(fooddish.FoodDishDirectory('data'))


########################################################################
## Unit Testing                                                       ##
########################################################################
if __name__ == "__main__":

    import unittest
        
    ########################################################################
    class TestStringWrapper(AssetBase):
        def __init__(self, string):
            AssetBase.__init__(self)
            self.string = string

    ########################################################################
    class TestCheekPouch(GenericAssetPouch):

        def __init__(self):
            GenericAssetPouch.__init__(self, TestStringWrapper, TestStringWrapper("BrokenReference"))
            self._ValidAssetList = {"ValidReference":TestStringWrapper("TheValidReference"), "SubReference":TestStringWrapper("TheSubReference")}
            
        def _load(self, assetid):
            if (self._ValidAssetList.has_key(assetid)):
                return self._ValidAssetList[assetid]
            return None
                        
        def AddAsset(self, ID, value):
            self._ValidAssetList[ID] = value
            
        def DelAsset(self, ID):
            del self._loaded_assets[ID]
            del self._ValidAssetList[ID]
        
        def AvailableAssetIDs(self):
            return self._ValidAssetList.keys()
            

    ########################################################################
    class ObjectWithReference(object):
        ref = AssetRef(TestStringWrapper, 'default')
        
        def __init__(self, assetid):
            self.ref = assetid
        
    ########################################################################
    RegisteredPouch = TestCheekPouch()
    RegisterPouch(RegisteredPouch)

    ########################################################################
    class FirstAssetClass(AssetBase):
        ReflectedFloat = AssetFloat(1.0)
        ReflectedInt = AssetInteger(1)
        
    ########################################################################
    class SecondAssetClass(FirstAssetClass):
        ReflectedString = AssetString("default")
        Reference = AssetRef(TestStringWrapper, "ValidReference")

    ########################################################################
    class UnitTestCP(unittest.TestCase):
        def setUp(self):
            pass
        
        def testRegistration(self):
            self.assert_(TestStringWrapper in AvailableAssetClassTypes())
    
        def testDisgorging(self):
            self.assert_(Disgorge(TestStringWrapper, "ValidReference").string == "TheValidReference")
    
        def testBrokenReference(self):
            obj = ObjectWithReference("TestBroken")
            self.assert_(obj.ref.string == "BrokenReference")
            RegisteredPouch.AddAsset("TestBroken", TestStringWrapper("NoLongerBroken"))
            self.assert_(obj.ref.string == "NoLongerBroken")
            RegisteredPouch.DelAsset("TestBroken")

        def testGetAssetID(self):
            obj = ObjectWithReference("default")
            self.assert_(type(obj).ref.get_assetid(obj) == 'default')

        def testAssetSerialization(self):
            obj = SecondAssetClass()
            obj.ReflectedFloat = 5.0
            obj.ReflectedString = "NotDefault"
            obj.NotReflectedFlag = False
            self.assertEqual(obj.ReflectedInt, 1)
            pickled = pickle.dumps(obj)
            loaded = pickle.loads(pickled)
            self.assertEqual(loaded.ReflectedFloat, obj.ReflectedFloat)
            self.assertEqual(loaded.ReflectedString, obj.ReflectedString)
            self.assertEqual(loaded.ReflectedInt, 1)
            self.assert_(not hasattr(loaded, "NotReflectedFlag"))
            del loaded.ReflectedString
            self.assertEqual(loaded.ReflectedString, "default")
                        
        def testAssetReference(self):
            obj = SecondAssetClass()
            self.assertEqual(obj.__class__.Reference.get_assetid(obj), "ValidReference")
            self.assertEqual(obj.Reference.string, "TheValidReference")
            obj.Reference = "SubReference"
            self.assertEqual(obj.Reference.string, "TheSubReference")
            del obj.Reference
            self.assertEqual(obj.__class__.Reference.get_assetid(obj), "ValidReference")
            self.assertEqual(obj.Reference.string, "TheValidReference")

        def testOnlyOneVec2DType(self):
            import motherhamster as mh
            self.assert_(vec2d == mh.mathobj.vec2d)
                        
    ########################################################################

    unittest.main()
########################################################################
