"""
a little class to take care of communicating to/from editor processes
it uses sockets & select for asynchronous operation 
(it avoids asyncore because we don't want to add to the global asyncore map
 because games might want to use asyncore and we shouldn't be in the way)
 
(C) Copyright 2007 Hamster Republic Productions
"""

########################################################################
import socket
import select
import marshal
from motherhamster.enums import Enum

########################################################################
class Connection(object):
    
    states = Enum("no_connection", "listening", "connecting", "connected")
    def __init__(self, connection_callback, error_callback):
        self.socket = None
        self.state = Connection.states.no_connection
        self.input_buffer = ""
        self.connection_callback = connection_callback
        self.error_callback = error_callback
        self.message_handlers = {}

    def set_message_handler(self, msg, callback):
        self.message_handlers[msg] = callback
        
    def listen(self, port):
        assert(self.state == Connection.states.no_connection)
        self.state = Connection.states.listening
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(0)
        self.socket.bind(('', port))
        self.socket.listen(1)
    
    def connect(self, address, port):
        assert(self.state == Connection.states.no_connection)
        self.state = Connection.states.connecting
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(0)
        errno = self.socket.connect_ex((address, port))
    
    def update(self):
        if self.state == Connection.states.listening:
            socket_list = [self.socket]
            readable, writable, errored = select.select(socket_list, [], socket_list, 0)
            if self.socket in readable:
                self.socket, address = self.socket.accept()
                self.socket.settimeout(0)
                self.state = Connection.states.connected
                self.connection_callback(address)
            if self.socket in errored:
                self.error_callback("ERROR: failure listening for connection")
                self.disconnect()
        elif self.state == Connection.states.connecting:
            socket_list = [self.socket]
            readable, writable, errored = select.select([], socket_list, socket_list, 0)
            if self.socket in writable:
                self.state = Connection.states.connected
                self.socket.settimeout(0)
                self.connection_callback(self.socket.getpeername())
            if self.socket in errored:
                self.error_callback("ERROR: failure connecting to remote address")
                self.disconnect()

        if self.state == Connection.states.connected:
            check_for_data = True
            while check_for_data:
                socket_list = [self.socket]
                readable, writable, errored = select.select(socket_list, [], socket_list, 0)
                if self.socket in readable:
                    check_for_data = self.recieve_message()
                else:
                    check_for_data = False
                    if self.socket in errored:
                        self.error_callback("ERROR: unknown problem with socket, closing")
                        self.disconnect()
    
    def disconnect(self):
        if self.state != Connection.states.no_connection:
            self.socket.close()
            self.input_buffer = ""
            self.socket = None
            self.connection_callback(None)
            self.state = Connection.states.no_connection

    def recieve_message(self):
        assert(self.state == Connection.states.connected)
        self.socket.settimeout(0)
        msg = self.socket.recv(4096)
        if len(msg) > 0:
            self.input_buffer += msg
            if len(self.input_buffer) > 11:
                assert(self.input_buffer[8] == 'x')
                packet_len = int(self.input_buffer[0:8], 16)
                if len(self.input_buffer) >= packet_len:
                    msg_len = int(self.input_buffer[9:11], 16)
                    msg = self.input_buffer[11:(11 + msg_len)]
                    arg_string = self.input_buffer[(11 + msg_len):packet_len]
                    self.input_buffer = self.input_buffer[packet_len:]
                    arg = marshal.loads(arg_string)
                    self.process_message(msg, arg)
            return True
        return False

    def process_message(self, msg, arg):
        if msg in self.message_handlers:
            self.message_handlers[msg](arg)
        else:
            self.error_callback('ERROR: Unknown message "'+msg+'" recieved')
        
    def send_message(self, msg, arg):
        assert(self.state == Connection.states.connected)
        arg_string = marshal.dumps(arg)
        packet_len = len(msg) + len(arg_string) + 11
        packet = "%08xx%02x%s%s" % (packet_len, len(msg), msg, arg_string)
        sent_bytes = self.socket.send(packet)
        assert(sent_bytes == len(packet))
        
########################################################################

if __name__ == "__main__":
    import unittest
    import threading
    import time

    ####################################################################
    class TestConnection(unittest.TestCase):
    
        def server_connected(self, address):
            print "server connected:",address
            
        def client_connected(self, address):
            print "client connected:",address

        def server_hello(self, arg):
            print "recieved hello"
            self.assert_(arg == "world")
            self.hello_recieved = True

        def client_goodbye(self, arg):
            print "recieved goodbye"
            self.assert_(arg == None)
            self.goodbye_recieved = True
            
        def connection_error(self, error):
            print "connection error:",error
            self.last_error = error
            
        def _client_process(self):
            client = Connection(self.client_connected, self.connection_error)
            client.connect('localhost', 39731)
            self.assert_(client.state == Connection.states.connecting)
            client.set_message_handler("hello", self.server_hello)
            while self.hello_recieved == False and time.time() < self.timeout_time:
                client.update()
            client.send_message("goodbye", None)
            client.disconnect()

        def testConnection(self):
            print "starting our test"
            self.timeout_time = time.time() + 3
            self.hello_recieved = False
            self.goodbye_recieved = False
            self.last_error = None
            client_thread = threading.Thread(name="ClientThread", target=self._client_process)
            client_thread.start()
            
            server = Connection(self.server_connected, self.connection_error)
            server.set_message_handler("goodbye", self.client_goodbye)
            server.listen(39731)
            self.assert_(server.state == Connection.states.listening)
            while server.state != Connection.states.connected and time.time() < self.timeout_time:
                server.update()
            server.send_message("hello", "world")
            while self.goodbye_recieved == False and time.time() < self.timeout_time:
                server.update()
            server.disconnect()
            self.assert_(self.hello_recieved)
            self.assert_(self.goodbye_recieved)
            
    ####################################################################
    unittest.main()
            