"""
The standard framework for a game that uses the MotherHamster Engine

Games should derive their main class from HamsterWheel, and should then
call Run() on their main class in order to start the game loop

(C) Copyright 2004 Hamster Republic Productions
"""

########################################################################

import pygame
#TODO: figure out how to get some simple loader window up very quickly...
import os
import weakref

from motherhamster.mathobj import vec2d
from motherhamster.enums import Enum
from motherhamster import profiler

########################################################################
class _MouseData:
    """ a container class for mouse data in the MotherHamster framework
    
        it's purpose is to lock it's state to individual update calls
        one design goal is to preserve as much information about the
        mouse events as possible, so in addition to storing state, it
        also preserves the click locations"""
    
    class _MousePath(list):
        """ a class to contain a drag path for mouse movement
            the idea is to log the mouse movement 
            as precisely as possible for gestural or drawing apps"""
        def __init__(self, left_down, right_down, center_down):
            list.__init__(self)
            self.left_down = left_down
            self.right_down = right_down
            self.center_down = center_down
            
    def _translate(self, pos):
        return vec2d(pos[0], self.maxy - pos[1])
    
    def __init__(self, width, height):
        self.maxy = height - 1
        self.pos = self._translate(pygame.mouse.get_pos())
        self.last_pos = self.pos
        self.scroll_wheel = 0
        self.last_scroll_wheel = self.scroll_wheel
        self.left_clicks = []
        self.right_clicks = []
        self.center_clicks = []
        self.left_down = False
        self.right_down = False
        self.center_down = False
        self.log = []
        self.__pending_log = []
    
    def set_screen_size(self, width, height):    
        self.maxy = height - 1
        
    def scroll_change(self):
        """ gets the change in the scroll wheel between update calls """
        return self.scroll_wheel - self.last_scroll_wheel
    
    def _preupdate(self):
        """ called by HamsterWheel before processing events for an update """
        self.last_pos = self.pos
        self.last_scroll_wheel = self.scroll_wheel
        del self.left_clicks[:]
        del self.right_clicks[:]
        del self.center_clicks[:]
        self.log = self.__pending_log
        self.__pending_log = []
        
    def _event_motion(self, event):
        """ called by HamsterWheel for a pygame event """
        # CHECK: is there anything we can do to get great drag support? like maybe track relative drag position? I dunno...
        self.pos = self._translate(event.pos)
        self.left_down = (event.buttons[0] != 0)
        self.center_down = (event.buttons[1] != 0)
        self.right_down = (event.buttons[2] != 0)
        self._log_motion(self.pos, self.left_down, self.center_down, self.right_down)
        
    def _event_down(self, event):
        """ called by HamsterWheel for a pygame event """
        if event.button == 1:
            self.left_clicks.append(self._translate(event.pos))
            self.left_down = True
        elif event.button == 2:
            self.center_clicks.append(self._translate(event.pos))
            self.center_down = True
        elif event.button == 3:
            self.right_clicks.append(self._translate(event.pos))
            self.right_down = True
        elif event.button == 4:
            self.scroll_wheel -= 1
        elif event.button == 5:
            self.scroll_wheel += 1
        self._log_motion(self.pos, self.left_down, self.center_down, self.right_down)
        
    def _event_up(self, event):
        """ called by HamsterWheel for a pygame event """
        if event.button == 1:
            self.left_down = False
        elif event.button == 2:
            self.center_down = False
        elif event.button == 3:
            self.right_down = False
        self._log_motion(self.pos, self.left_down, self.center_down, self.right_down)
    
    def _log_motion(self, pos, left_down, right_down, center_down):
        if len(self.__pending_log) == 0:
            path = _MouseData._MousePath(left_down, right_down, center_down)
            path.append(pos)
            self.__pending_log.append(path)
        else:
            path = self.__pending_log[-1]
            if path.left_down != left_down or path.right_down != right_down or path.center_down != center_down:
                path.append(pos)
                path = _MouseData._MousePath(left_down, right_down, center_down)
                path.append(pos)
                self.__pending_log.append(path)
            else:
                if path[-1] != pos:
                    path.append(pos)
            

########################################################################
class _KeyData:
    """ a container class for mouse data in the MotherHamster framework    
        it's purpose is to lock it's state to individual update calls"""
        
    def __init__(self):
        self.__held = {}
        self.__pressed = {}

    def holding(self, keycode):
        """ tells if the key corresponding to the pygame keycode should
            be considered held for this update """
        if keycode in self.__held:
            return self.__held[keycode]
        return False

    def pressed(self, keycode):
        """ returns the number of times the key corresponding to the given
            pygame keycode was pressed for this update """
        if keycode in self.__pressed:
            return self.__pressed[keycode]
        return 0
        
    def _preupdate(self):
        """ called by HamsterWheel before processing events for an update """
        self.__pressed.clear()
    
    def _event_down(self, event):
        """ called by HamsterWheel for a pygame event """
        # update key state
        press_count = 1
        if event.key in self.__pressed:
            press_count = self.__pressed[event.key] + 1
        self.__pressed[event.key] = press_count
        self.__held[event.key] = True
    
    def _event_up(self, event):
        """ called by HamsterWheel for a pygame event """
        if event.key in self.__held:
            del self.__held[event.key]
        
########################################################################
class HamsterWheel:
    """ The base class for a game in the MotherHamster framework
    
        A game class can set it's screen_width, screen_height,
        updates_per_second, max_frame_skip to customize itself.
        
        To run your game, you should call run_game on an instance of
        your class. Other modules may use HamsterWheel.the_game to
        access the instance of the game class currently running
        
        modules may access the screen, fullscreen, run_time & is_playing
        attributes of the instance of the game"""

    the_game = None
    events = Enum("FrameStart", "FrameEnd", "UpdateStart", "UpdateEnd", "DrawStart", "DrawEnd")

    #---------------------------------------------
    def __init__(self, title = "HamsterWheel"):
        # provide static access to the game object
        HamsterWheel.the_game = self
        pygame.init()
        #CHECK: I know we'll want to do more with the mixer - crossfading, ambient loop tracks, etc.
        try:
            pygame.mixer.init()
        except:
            #CHECK: this should be more user visible - like a message box
            print "WARNING: Sound Support disabled: %s" % pygame.get_error()
            pass
        # init the public stuff
        self.runtime = .0
        self.is_playing = False
        self.title = title
        # CHECK: should pull this from a config file for the user?
        if (not hasattr(self, 'fullscreen')):
            self.fullscreen = False
        #from displaySDL import Display
        from displayGL import Display
        self.display = Display()
        # setting defaults for various settings, derived class should override these
        if (not hasattr(self, 'screen_width')):
            self.screen_width = 800
        if (not hasattr(self, 'screen_height')):
            self.screen_height = 600
        if (not hasattr(self, 'updates_per_second')):
            self.updates_per_second = 60
        if (not hasattr(self, 'max_frame_skip')):
            self.max_frame_skip = 6
        if (not hasattr(self, 'grab_mouse')):
            self.grab_mouse = True
        if (not hasattr(self, 'hide_system_mouse')):
            self.hide_system_mouse = True
        # now we set all the variables for interfacing with the editor
        self.game_clock_speed = 1.0
        self.mouse = _MouseData(self.screen_width, self.screen_height)
        self.keys = _KeyData()
        self.__subscribers = []
        self.active = True
        self.editor = None
        self._update_profile = profiler.RegisterFunction(self, "_update")
        self._draw_profile = profiler.RegisterFunction(self, "_draw")
        self._flip_profile = profiler.RegisterFunction(self, "_flip")
        self._wait_profile = profiler.RegisterFunction(pygame.time, "wait", "pygame.time")
        self._draw_fps_profile = profiler.RegisterFunction(self, "_draw_fps")
        self.show_fps = False
        self.draw_debug_enabled = False
                    
    #---------------------------------------------
    def __del__(self):
        # clean up pygame
        pygame.quit()
        pygame.mixer.quit()
        del self.__subscribers

    #---------------------------------------------
    def run_game(self):
        """ The main game loop, should be called on a derived class to run the game.
            this routine controls synchronization with the editor & the game clock """
        self.is_playing = True
        
        self.__set_screen_mode()
        self.__process_events()
        self.display.erase()
        self.display.flip()
        
        msperupdate = 1000.0/self.updates_per_second
        time_bucket = 0.0
        pygame.mouse.set_visible(not self.hide_system_mouse)
        if (self.fullscreen):
            self.grab_mouse = True
        #CHECK: on windows we may not start activated - but we aren't told whether we are... ARG
        # we don't want to grab the mouse if our window is hidden, so what is the right thing to do?
        pygame.event.set_grab(self.grab_mouse)
        self.__report_event(self.events.FrameStart)
        last_clock_ticks = pygame.time.get_ticks()
        while self.is_playing:
            # locally cache clock speed, cause editor edits asynchronously
            current_clock_speed = self.game_clock_speed
            # find out the time elapsed, delaying if necessary, scaling by speed
            if (current_clock_speed > 0):
                new_clock_ticks = pygame.time.get_ticks()
                time_elapsed = current_clock_speed*(new_clock_ticks - last_clock_ticks)
                time_needed_for_update = int((msperupdate - time_bucket - time_elapsed)/current_clock_speed)
                if time_needed_for_update > 0:
                    pygame.time.wait(time_needed_for_update)
                    new_clock_ticks = pygame.time.get_ticks()
                    time_elapsed = current_clock_speed*(new_clock_ticks - last_clock_ticks)
                time_bucket += time_elapsed
                self.runtime += time_elapsed
                last_clock_ticks = new_clock_ticks
            else:
                # if the game is paused, get events & tick to make sure the time bucket doesn't fill
                self.__process_events()
                time_bucket = 0.0
            updates_this_draw = 0
            # do the right number of updates based on time elapsed
            while (time_bucket >= msperupdate):
                if (updates_this_draw >= self.max_frame_skip):
                    time_bucket = time_bucket % msperupdate
                else:
                    updates_this_draw = updates_this_draw + 1
                    time_bucket -= msperupdate
                    self.__process_events()
                    self._update(msperupdate/1000.0)
            # draw if we had an update, or if paused
            if (updates_this_draw > 0 or self.game_clock_speed <= 0):
                self._draw()
                if self.show_fps:
                    self._draw_fps()
                self._flip()
                self.__report_event(self.events.FrameEnd)
                profiler.AdvanceFrame()
                self.__report_event(self.events.FrameStart)
        self.__report_event(self.events.FrameEnd)

    #---------------------------------------------
    def __set_screen_mode(self):
        # CHECK: what about icon?
        self.display.set_mode(self.screen_width, self.screen_height, self.fullscreen)
        self.mouse.set_screen_size(self.screen_width, self.screen_height)
        self.__set_caption()

    #---------------------------------------------
    def __set_caption(self):
        window_title = self.title
        if self.grab_mouse and self.active:
            window_title += " (mouse locked to window)"
        pygame.display.set_caption(window_title, self.title)
        
    #---------------------------------------------
    def __process_events(self):
        # let the input state stuff prep for the next frame
        self.mouse._preupdate()
        self.keys._preupdate()
        # event handling
        for event in pygame.event.get():
            self.process_sdl_event(event)
        
    #---------------------------------------------
    def _update(self, seconds):
        self.__report_event(self.events.UpdateStart)
        self.update(seconds)
        self.__report_event(self.events.UpdateEnd)
        
    #---------------------------------------------
    def _draw(self):
        self.__report_event(self.events.DrawStart)
        self.draw()
        self.__report_event(self.events.DrawEnd)
        if self.draw_debug_enabled:
            self.draw_debug()
    
    #---------------------------------------------
    def _draw_fps(self):
        average_count = 6
        draw_time = self._draw_profile.total_avg(average_count)
        update_time = self._update_profile.total_avg(average_count)
        flip_time = self._flip_profile.total_avg(average_count)
        wait_time = self._wait_profile.total_avg(average_count)
        fps_time = self._draw_fps_profile.total_avg(average_count)
        total_time = profiler.FrameTimeAverage(average_count)
        length_time = self.display.width*15.0
 
        self.display.rect([0,9], [update_time*length_time, 12], (1,0,0,1))        
        self.display.rect([0,6], [draw_time*length_time, 9], (0,0,1,1))
        self.display.rect([0,3], [flip_time*length_time, 6], (0,1,0,1))
        frame_time = update_time + draw_time + flip_time
        self.display.rect([0,0], [frame_time*length_time, 3], (1,1,1,1))
        frame_time_end = frame_time + wait_time
        self.display.rect([frame_time*length_time,0], [frame_time_end*length_time, 3], (.5,.5,.5,1))
        frame_time = frame_time_end
        frame_time_end = frame_time + fps_time
        self.display.rect([frame_time*length_time,0], [frame_time_end*length_time, 3], (1,0,1,1))
        frame_time = frame_time_end
        frame_time_end = total_time
        self.display.rect([frame_time*length_time,0], [frame_time_end*length_time, 3], (1,1,0,1))

        frame_div = 60*16
        while frame_div > 15:
            marker_x = int(length_time/frame_div)
            if frame_div == 60:
                self.display.rect([marker_x - 1, 0], [marker_x + 1, 15], (1,.3,1,1))
            else:
                self.display.rect([marker_x - 1, 0], [marker_x + 1, 14], (.6,0,.6,1))
            frame_div /= 2       
        
    #---------------------------------------------
    def _flip(self):
        self.display.flip()

    #----- functions that should be overriden by the base class
    def update(self, seconds):
        """ Should be defined in a derived class to update the game state
            seconds is the amount of game time that should pass"""
        print "HamsterWheel update (elapsed:" + str(seconds) + ")"
    
    #---------------------------------------------
    def draw(self):
        """ Should be defined in a derived class to draw the game"""
        print "HamsterWheel draw"
    
    #---------------------------------------------
    def draw_debug(self):
        """ Should be defined in a derived class to draw the game"""
        print "HamsterWheel draw debug"
    
    #---------------------------------------------
    def process_sdl_event(self, event):
        """ derived classes should call the base classes process function """
        if event.type == pygame.QUIT:
            self.quit()
            
        elif event.type == pygame.ACTIVEEVENT:
            # CHECK: we should make games not run when they are not active, what's the best way to do that?
            #  simple state flag? stop calling update? call idle_update or some crazy thing like that?
            if (event.state & 2):
                if (event.gain):
                    self.active = True
                    # full screen apps should always grab the mouse, for multi-monitor games
                    pygame.event.set_grab(self.grab_mouse or self.fullscreen)
                else:
                    self.active = False
                    pygame.event.set_grab(False)
                self.__set_caption()
            
        elif event.type == pygame.KEYDOWN:
            self.keys._event_down(event)
            # do the full screen stuff
            if event.key == pygame.K_RETURN and event.mod & pygame.KMOD_ALT:
                self.fullscreen = not self.fullscreen
                self.__set_screen_mode()
            # pop out an editor window
            if event.key == pygame.K_BACKQUOTE:
                # am kicking out of fullscreen for editor, to make sure editor window is shown
                if (self.fullscreen):
                    self.fullscreen = False
                    self.__set_screen_mode()
                self.grab_mouse = False
                pygame.event.set_grab(self.grab_mouse)
                import motherhamster.editor
                self.editor = motherhamster.editor
                self.editor.SpawnEditor()
            if event.key == pygame.K_f and event.mod & pygame.KMOD_ALT:
                self.show_fps = not self.show_fps
                self._update_profile.activate()
                self._draw_profile.activate()
                self._flip_profile.activate()
                self._wait_profile.activate()
                self._draw_fps_profile.activate()
            if event.key == pygame.K_F4 and event.mod & pygame.KMOD_ALT:
                # force alt-f4 to be quit on windows
                if (os.name == 'nt'):                
                    pygame.event.post(pygame.event.Event(pygame.QUIT))
                                    
        elif event.type == pygame.KEYUP:
            self.keys._event_up(event)
            
        # Mouse processing
        elif event.type == pygame.MOUSEMOTION:
            self.mouse._event_motion(event)
        elif event.type == pygame.MOUSEBUTTONDOWN:
            self.mouse._event_down(event)
        elif event.type == pygame.MOUSEBUTTONUP:
            self.mouse._event_up(event)
                
    #---------------------------------------------
    def quit(self):
        """ called when a quit message is posted
        
            the HamsterWheel base class responds by setting is_playing to false
            in order to end the update a loop. derived classes should call the base
            class, but may conditionally ignore quit messages by not doing so """
        if self.editor:
            self.editor.KillEditor()
        self.is_playing = False

    #---------------------------------------------
    def register_for_events(self, obj):
        """ call this with self to get your function "hamsterwheel_event"
            called before and after updates, draws and frames
            hamsterwheel_event will be called with one arg eventtype which
            will be one of the HamsterWheel.event members """
        for weak_ref in self.__subscribers:
            if weak_ref() == obj:
                return
        self.__subscribers.append(weakref.ref(obj, self.unregister_for_events))
        
    #---------------------------------------------
    def unregister_for_events(self, obj):
        """ called to stop getting event calls """
        if self != None:
            for weak_ref in self.__subscribers:
                if (weak_ref == obj or weak_ref() == obj):
                    self.__subscribers.remove(weak_ref)
                    return
        
    #---------------------------------------------
    def __report_event(self, eventtype):
        for subscriber in self.__subscribers:
            subscriber().hamsterwheel_event(eventtype)
            
########################################################################

if __name__ == '__main__':
    import unittest
    HamsterWheel()
    
    ####################################################################
    class UnitTestHW(unittest.TestCase):
        def setup(self):
            self.last_event = None
            
        def hamsterwheel_event(self, eventtype):
            self.last_event = eventtype
            
        def testEventRegistration(self):
            HamsterWheel.the_game.register_for_events(self)
            HamsterWheel.the_game._HamsterWheel__report_event(HamsterWheel.events.FrameStart)
            self.assert_(self.last_event == HamsterWheel.events.FrameStart)
            HamsterWheel.the_game.unregister_for_events(self)
            HamsterWheel.the_game._HamsterWheel__report_event(HamsterWheel.events.FrameEnd)
            self.assert_(self.last_event == HamsterWheel.events.FrameStart)            

    ########################################################################

    unittest.main()

    ########################################################################
