"""
a module that manages ad-hoc performance monitoring 
(C) Copyright 2005 Hamster Republic Productions
"""

from array import array
import types
import time
import platform

########################################################################
if 'Windows' in platform.platform(aliased = True, terse = True):
    _TimeFunc = time.clock
else:
    _TimeFunc = time.time
FunctionProfiles = []

########################################################################
class FunctionTracker(object):
    """Tracks calls and time in calls to a function"""
    _NestedTime = 0
    SampleCount = 5*60
    Index = 0

    def __init__(self, parent, function_name, parent_name = None):
        assert(callable(getattr(parent, function_name)))
        self.parent_obj = parent
        self.function_name = function_name
        self.call_counts = array('f', [0]*FunctionTracker.SampleCount)
        self.nested_times = array('f', [0]*FunctionTracker.SampleCount)
        self.owned_times = array('f', [0]*FunctionTracker.SampleCount)
        self.total_time = 0
        self.frame_max = 0
        self.frame_total = 0
        self.old_func = None
        if parent_name != None:
            self.__dict__["object_name"] = parent_name

    class ParentObjNameProperty(object):
        def __get__(self, obj, objtype=None):    #IGNORE:W0613
            if (isinstance(obj.parent_obj, types.TypeType)):
                name = repr(obj.parent_obj).strip("<'>").split('.')[-1].split(' at ')[0]
            else:
                type_str = repr(obj.parent_obj).strip("<';>").split(' ')[0].split('.')[-1]
                name = type_str + " : instance"
            obj.__dict__["object_name"] = name
            return name
    object_name = ParentObjNameProperty()
    
    def activate(self):
        if (self.old_func == None):
            self.old_func = getattr(self.parent_obj, self.function_name)
            setattr(self.parent_obj, self.function_name, self)
            
    def deactivate(self):
        if (self.old_func != None):
            setattr(self.parent_obj, self.function_name, self.old_func)
            self.old_func = None

    def is_active(self):
        return self.old_func != None
    
    def __get__(self, obj, objtype=None):
        return types.MethodType(self, obj, objtype)
       
    def __call__(self, *args, **kwargs):
        #CHECK: this nested time stuff does not work testing multithreaded stuff :(
        # record the start
        OldNestedTime = FunctionTracker._NestedTime
        FunctionTracker._NestedTime = .0
        self.call_counts[FunctionTracker.Index] += 1.0
        # call the function
        start = _TimeFunc()
        retval = self.old_func(*args, **kwargs)
        end_time = _TimeFunc()
        # record the stop
        total = end_time - start
        nested = FunctionTracker._NestedTime
        self.owned_times[FunctionTracker.Index] += total - nested
        self.nested_times[FunctionTracker.Index] += nested
        self.frame_total += total
        FunctionTracker._NestedTime = OldNestedTime + total
        return retval
        
    def advance(self):
        self.call_counts[FunctionTracker.Index] = 0
        self.nested_times[FunctionTracker.Index] = 0
        self.owned_times[FunctionTracker.Index] = 0
        self.total_time += self.frame_total
        self.frame_max = max(self.frame_max, self.frame_total)
        self.frame_total = 0
        
    def calls(self, index):
        return self.call_counts[index % FunctionTracker.SampleCount]
        
    def nested(self, index):
        return self.nested_times[index % FunctionTracker.SampleCount]
        
    def owned(self, index):
        return self.owned_times[index % FunctionTracker.SampleCount]
            
    def call_avg(self, count):
        sum = .0
        end = FunctionTracker.Index
        count = min(count, FunctionTracker.SampleCount - 1)
        start = end - count
        if start < 0:
            start = (end - count) % FunctionTracker.SampleCount
            for i in xrange(start, FunctionTracker.SampleCount):
                sum += self.call_counts[i]
            start = 0
        for i in xrange(start, end):
            sum += self.call_counts[i]
        return sum/count
        
    def nested_avg(self, count):
        sum = .0
        end = FunctionTracker.Index
        count = min(count, FunctionTracker.SampleCount - 1)
        start = end - count
        if start < 0:
            start = (end - count) % FunctionTracker.SampleCount
            for i in xrange(start, FunctionTracker.SampleCount):
                sum += self.nested_times[i]
            start = 0
        for i in xrange(start, end):
            sum += self.nested_times[i]
        return sum/count
        
    def owned_avg(self, count):
        sum = .0
        end = FunctionTracker.Index
        count = min(count, FunctionTracker.SampleCount - 1)
        start = end - count
        if start < 0:
            start = (end - count) % FunctionTracker.SampleCount
            for i in xrange(start, FunctionTracker.SampleCount):
                sum += self.owned_times[i]
            start = 0
        for i in xrange(start, end):
            sum += self.owned_times[i]
        return sum/count

    def total_avg(self, count):
        sum = .0
        end = FunctionTracker.Index
        count = min(count, FunctionTracker.SampleCount - 1)
        start = end - count
        if start < 0:
            start = (end - count) % FunctionTracker.SampleCount
            for i in xrange(start, FunctionTracker.SampleCount):
                sum += self.nested_times[i]
                sum += self.owned_times[i]
            start = 0
        for i in xrange(start, end):
            sum += self.nested_times[i]
            sum += self.owned_times[i]
        return sum/count
        
            
########################################################################
def RegisterFunction(parent, function_name, parent_name = None):
    for profile in FunctionProfiles:
        if profile.parent_obj == parent and profile.function_name == function_name:
            return profile
    profile = FunctionTracker(parent, function_name, parent_name)
    FunctionProfiles.append(profile)
    return profile

########################################################################
def UnregisterFunction(parent, function_name):
    for profile in FunctionProfiles:
        if profile.parent_obj == parent and profile.function_name == function_name:
            profile.deactivate()
            FunctionProfiles.remove(profile)
            return True
    return False

########################################################################
def IsFunctionRegistered(parent, function_name):
    for profile in FunctionProfiles:
        if profile.parent_obj == parent and profile.function_name == function_name:
            return True
    return False
    
########################################################################
_FrameTimes = array('f', [0]*FunctionTracker.SampleCount)
_FrameStart = _TimeFunc()

def    AdvanceFrame():
    global _FrameStart
    new_time = _TimeFunc()
    _FrameTimes[FunctionTracker.Index] = new_time - _FrameStart
    _FrameStart = new_time
    FunctionTracker.Index = (FunctionTracker.Index + 1) % FunctionTracker.SampleCount
    for profile in FunctionProfiles:
        profile.advance()

########################################################################
def FrameTime(index):
    return _FrameTimes[index % FunctionTracker.SampleCount]    

########################################################################
def FrameTimeAverage(count):
    sum = .0
    end = FunctionTracker.Index
    count = min(count, FunctionTracker.SampleCount - 1)
    start = end - count
    if start < 0:
        start = (end - count) % FunctionTracker.SampleCount
        for i in xrange(start, FunctionTracker.SampleCount):
            sum += _FrameTimes[i]
        start = 0
    for i in xrange(start, end):
        sum += _FrameTimes[i]
    return sum/count
        
########################################################################
# Unit Testing
########################################################################
if __name__ == '__main__':
    import unittest
    import __main__

    def DummyFunc(self):
        pass
        
    ####################################################################
    class UnitTestProfiler(unittest.TestCase):

        def RecursiveWaitThreeSeconds(self):
            self.WaitOneSecond()
            time.sleep(1)
            self.WaitOneSecond()
            
        def WaitOneSecond(self):
            time.sleep(1)

        def ThirdLevelCallFunction(self):
            self.RecursiveWaitThreeSeconds()
            
        def testRegistration(self):
            self.assert_(not IsFunctionRegistered(__main__, "RegisterFunction"))
            rfTracker = RegisterFunction(__main__, "RegisterFunction")
            self.assert_(IsFunctionRegistered(__main__, "RegisterFunction"))
            self.assert_(UnregisterFunction(__main__, "RegisterFunction"))
            self.assert_(not IsFunctionRegistered(__main__, "RegisterFunction"))
            self.assert_(not UnregisterFunction(__main__, "RegisterFunction"))
            
        def testTimingAndCalls(self):
            print 'Doing Timing Tests...'
            RegisterFunction(__main__, "RegisterFunction").activate()
            RegisterFunction(UnitTestProfiler, "WaitOneSecond").activate()
            RegisterFunction(UnitTestProfiler, "RecursiveWaitThreeSeconds").activate()

            self.WaitOneSecond()
            self.RecursiveWaitThreeSeconds()

            AdvanceFrame()
            rfTracker = RegisterFunction(__main__, "RegisterFunction")
            wtsTracker = RegisterFunction(UnitTestProfiler, "RecursiveWaitThreeSeconds")
            wosTracker = RegisterFunction(UnitTestProfiler, "WaitOneSecond")
            
            rfTracker.deactivate()
            wosTracker.deactivate()
            
            self.RecursiveWaitThreeSeconds()
            self.assertEqual(rfTracker.calls(FunctionTracker.Index - 1), 2)
            self.assertEqual(rfTracker.calls(FunctionTracker.Index), 3)
            self.assertEqual(wtsTracker.calls(FunctionTracker.Index - 1), 1)
            self.assertAlmostEquals(wtsTracker.owned(FunctionTracker.Index - 1), 1, places=2)
            self.assertAlmostEquals(wtsTracker.nested(FunctionTracker.Index - 1), 2, places=2)
            self.assertEqual(wtsTracker.calls(FunctionTracker.Index), 1)
            self.assertAlmostEquals(wtsTracker.owned(FunctionTracker.Index), 3, places=2)
            self.assertEquals(wtsTracker.nested(FunctionTracker.Index), 0)
            self.assertEqual(wosTracker.calls(FunctionTracker.Index - 1), 3)
            self.assertAlmostEquals(wosTracker.owned(FunctionTracker.Index - 1), 3, places=1)
            self.assertEquals(wosTracker.nested(FunctionTracker.Index - 1), 0)
            self.assertEqual(wosTracker.calls(FunctionTracker.Index), 0)
            self.assertEquals(wosTracker.nested(FunctionTracker.Index), 0)
                        
            UnregisterFunction(__main__, "RegisterFunction")
            UnregisterFunction(UnitTestProfiler, "WaitOneSecond")
            UnregisterFunction(UnitTestProfiler, "RecursiveWaitThreeSeconds")

        def testTimerFunc(self):
            elapsedsum = .0
            samples = 100
            for i in range(samples):
                oldtime = _TimeFunc()
                newtime = oldtime
                while oldtime == newtime:
                    newtime = _TimeFunc()
                elapsedsum += newtime - oldtime
                time.sleep(.001)
            resolution = elapsedsum/samples
            frequency = 1/resolution
            print 'timer frequency: ', int(frequency), 'hertz'
            self.assert_(frequency > 60*100)
            
        def testAverageFunc(self):
            test_profile = RegisterFunction(__main__, "DummyFunc")
            for offset in xrange(0, FunctionTracker.SampleCount, 33):
                FunctionTracker.Index = offset
                for num in xrange(1, FunctionTracker.SampleCount):
                    test_profile.call_counts[FunctionTracker.Index] = num
                    test_profile.nested_times[FunctionTracker.Index] = num*2
                    test_profile.owned_times[FunctionTracker.Index] = num*3
                    _FrameTimes[FunctionTracker.Index] = num*10
                    FunctionTracker.Index = (FunctionTracker.Index + 1) % FunctionTracker.SampleCount
                    expected = (num+1)/2.0
                    self.assertEqual(test_profile.call_avg(num), expected)
                    self.assertEqual(test_profile.nested_avg(num), expected*2)
                    self.assertEqual(test_profile.owned_avg(num), expected*3)
                    self.assertEqual(test_profile.total_avg(num), expected*5)
                    self.assertEqual(FrameTimeAverage(num), expected*10)
            UnregisterFunction(__main__, "DummyFunc")
            
        def testThirdLevelStuff(self):
            third_level = RegisterFunction(UnitTestProfiler, "ThirdLevelCallFunction")
            second_level = RegisterFunction(UnitTestProfiler, "RecursiveWaitThreeSeconds")
            first_level = RegisterFunction(UnitTestProfiler, "WaitOneSecond")

            third_level.activate()
            second_level.activate()
            first_level.activate()
            
            self.ThirdLevelCallFunction()
            AdvanceFrame()

            self.assertAlmostEquals(third_level.owned(FunctionTracker.Index - 1), 0, places=2)
            self.assertAlmostEquals(second_level.owned(FunctionTracker.Index - 1), 1, places=2)
            self.assertAlmostEquals(first_level.owned(FunctionTracker.Index - 1), 2, places=2)

            self.assertAlmostEquals(third_level.nested(FunctionTracker.Index - 1), 3, places=2)
            self.assertAlmostEquals(second_level.nested(FunctionTracker.Index - 1), 2, places=2)
            self.assertAlmostEquals(first_level.nested(FunctionTracker.Index - 1), 0, places=2)

            UnregisterFunction(UnitTestProfiler, "ThirdLevelCallFunction")
            UnregisterFunction(UnitTestProfiler, "RecursiveWaitThreeSeconds")
            UnregisterFunction(UnitTestProfiler, "WaitOneSecond")

    ########################################################################

    unittest.main()
    